﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Runtime.InteropServices.WindowsRuntime;
using System.Threading.Tasks;

using Windows.ApplicationModel.DataTransfer;
using Windows.Devices.Input;
using Windows.Foundation;
using Windows.Foundation.Collections;
using Windows.Graphics.Imaging;
using Windows.UI.Input;
using Windows.UI.Input.Inking;
using Windows.Storage;
using Windows.Storage.Streams;
using Windows.Storage.Provider;
using Windows.UI;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Controls.Primitives;
using Windows.UI.Xaml.Data;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Media.Imaging;
using Windows.UI.Xaml.Navigation;
using Windows.UI.Xaml.Shapes;
using Windows.UI.Popups;

using Foxit.General;
using Foxit.General.PDF;

using Xfinium.Pdf;
using Xfinium.Pdf.Graphics;
using C1.Xaml.Pdf;


// The Blank Page item template is documented at http://go.microsoft.com/fwlink/?LinkId=234238

namespace PDFTest
{
    /// <summary>
    /// An empty page that can be used on its own or navigated to within a Frame.
    /// </summary>
    public sealed partial class MainPage : Page
    {
        DocHandle m_doc;
        PageHandle m_page;
        int m_curPageIndex = 0;
        int m_ActualSizeX = 0;
        int m_ActualSizeY = 0;
        Rect m_pageRect;
        Rect m_unionRect;
        PSIHandle m_hPSI;
        Classes.CCallBack m_CallBack = new Classes.CCallBack();
        PSIAppcallback m_appCallBack = new PSIAppcallback();
        PixelSource m_bitmap;
        WriteableBitmap m_wb;
        byte[] m_bytes;
        public Boolean m_bFirstStart = false;
        public String m_FileID;
        //used for share
        StorageFile m_pdfFile;
        private double m_screenHeight;

        //PDFViewer related variables
        private bool m_isTracing;
        private double? m_previousX;
        private double? m_previousY;
        //form a struct and list of those to trace the lines drawn on the screen
        private struct ViewerPoint
        {
            public double X1;
            public double Y1;
            public double X2;
            public double Y2;
        }
        private List<ViewerPoint> ViewerPointsList = new List<ViewerPoint>();


        //Canvas related variables
        private uint m_PenId;
        private uint _touchID;
        private Point m_PreviousContactPoint;
        private Point currentContactPt;
        private double x1;
        private double y1;
        private double x2;
        private double y2;
        private Color m_CurrentDrawingColor = Colors.Black;
        private double m_CurrentDrawingSize = 4;
        private Color m_CurrentHighlightColor = Colors.Yellow;
        private double m_CurrentHighlightSize = 8;
        private string m_CurrentMode = "Ink";
        private bool m_IsRecognizing = false;
        InkManager m_InkManager = new InkManager();
        InkManager m_HighLightManager = new InkManager();

        //debug flags
        private bool isForComponentOne = true;
        private bool isForFoxIt = false;

        public InkManager CurrentManager
        {
            get
            {
                if (m_CurrentMode == "Ink") return m_InkManager;
                return m_HighLightManager;
            }
        }

        public MainPage()
        {
            this.InitializeComponent();

            //register canvas events
            inkCanvas.PointerPressed += new PointerEventHandler(inkCanvas_PointerPressed);
            inkCanvas.PointerMoved += new PointerEventHandler(inkCanvas_PointerMoved);
            inkCanvas.PointerReleased += new PointerEventHandler(inkCanvas_PointerReleased);
            inkCanvas.PointerExited += new PointerEventHandler(inkCanvas_PointerReleased);

            string fileName = "TestFile.pdf";
            if (isForFoxIt)
            {
                double windowHeight = Window.Current.Bounds.Height;
                double windowWidth = Window.Current.Bounds.Width;

                //no need to setup for share, this was in the sample
                //RegisterForShare();

                scroll_pdf.Height = windowHeight * 0.95;
                scroll_pdf.Width = windowWidth * 0.95;

                m_appCallBack.InvalidateFunc = new InvalidatePSIDelegate(m_CallBack.Invalidate);
                m_appCallBack.GetOpacityFunc = new GetOpacityDelegate(m_CallBack.GetOpacity);
                m_appCallBack.ClientData = this;
                OpenFileForFoxIt(fileName);
            }
            if (isForComponentOne)
            {
                //PDF Viewer TEST
                //string fileName = "TestFile.pdf";
                OpenFileForPDFViewer(fileName);
                //END PDFViewer TEST
            }
        }

        //destructor
        ~MainPage()
        {
            if (m_page.pointer != 0)
            {
                Document.ClosePage(m_page);
                m_page.pointer = 0;
            }
            if (m_doc.pointer != 0)
            {
                Document.Close(m_doc);
                m_page.pointer = 0;
            }
            if (m_hPSI.pointer != 0)
            {
                PSI.DestroyEnvironment(m_hPSI);
                m_hPSI.pointer = 0;
            }
        }

        #region "PDF Viewer and XFinium Test"

        private async void OpenFileForPDFViewer(string fileName)
        {
            StorageFolder folder = KnownFolders.DocumentsLibrary;
            StorageFile file = await folder.GetFileAsync(fileName);
            //STEP 1: Read and load a file into the program, using ComponentOne PDF
            //StorageFile file = await StorageFile.GetFileFromPathAsync(filePath);
            var randomAccessStream = await file.OpenReadAsync();
            Stream stream = randomAccessStream.AsStreamForRead();
            //var randomAccessStream = await file.OpenAsync(FileAccessMode.ReadWrite);
            //Stream stream = randomAccessStream.AsStreamForWrite();

            var memStream = new MemoryStream();
            await stream.CopyToAsync(memStream);
            memStream.Position = 0;
            Stream loadStream = memStream.AsRandomAccessStream().AsStreamForWrite();

            pdfViewer.RenderMode = C1.Xaml.PdfViewer.RenderMode.System;
            pdfViewer.LoadDocument(loadStream);

            ////STEP 2: Modify the loaded stream in-memory, using XFinium PDF
            //PdfFixedDocument doc = new PdfFixedDocument(loadStream);

            //if (doc != null && doc.Pages.Count > 0)
            //{
            //    var pdfPage = doc.Pages[0];
            //    DrawWatermarkOverPageContent(pdfPage);
            //}
            //PdfStandardFont helvetica = new PdfStandardFont(PdfStandardFontFace.Helvetica, 12);
            //CreateLineAnnotations(doc, helvetica);

            ////try creating a drawing
            //PdfPen blackPen = new PdfPen(new PdfRgbColor(100, 100, 0), 1);
            //doc.Pages[0].Graphics.DrawLine(blackPen, new PdfPoint(30, 200), new PdfPoint(60, 200));
            //doc.Pages[0].Graphics.SaveGraphicsState();


            ////STEP 3: Re-load the modified PDF into memory
            //doc.Save(loadStream);
            //loadStream.Position = 0;
            //pdfViewer.LoadDocument(loadStream);

            //STEP 4: Save the file in a different name - SKIP FOR NOW, THE FOLLOWING WORKS
            //StorageFile saveFile = await folder.CreateFileAsync("Test2PDF.pdf");
            //await pdfViewer.SaveDocumentAsync(saveFile);

        }

        private async void pdfViewer_PointerMoved(object sender, PointerRoutedEventArgs e)
        {
            Pointer ptr = e.Pointer;
            e.Handled = true;
            return;

            if (ptr.PointerDeviceType == PointerDeviceType.Mouse && m_isTracing)
            {
                PointerPoint ptrPoint = e.GetCurrentPoint(pdfViewer);

                if (ptrPoint.IsInContact)
                {
                    PdfPen blackPen = new PdfPen(new PdfRgbColor(0, 0, 0), 1);
                    //get the loaded document from the viewer by saving it into a stream
                    var memStream = new MemoryStream();
                    Stream saveStream = memStream.AsRandomAccessStream().AsStreamForWrite();
                    await pdfViewer.SaveDocumentAsync(saveStream);

                    //get the page no
                    int pageNumber = pdfViewer.PageNumber;
                    //get the current coordinates
                    double currentX = ptrPoint.Position.X;
                    double currentY = ptrPoint.Position.Y;
                    if (m_previousX == null) { m_previousX = currentX; }
                    if (m_previousY == null) { m_previousY = currentY; }

                    System.Diagnostics.Debug.WriteLine("Current X:" + currentX.ToString() + "; Current Y:" + currentY.ToString());

                    //load stream into an editable document
                    PdfFixedDocument doc = new PdfFixedDocument(saveStream);
                    //get the selected page
                    if (pageNumber - 1 < 0) { pageNumber = 1; }
                    //get the scaling factor to convert position on page to position on the PDF page
                    double pageWidth = doc.Pages[pageNumber - 1].Width;
                    double pageHeight = doc.Pages[pageNumber - 1].Height;
                    double conversionFactorX = pageWidth / pdfViewer.ActualWidth;
                    double conversionFactorY = pageHeight / pdfViewer.ActualHeight;
                    //draw a line on the selected page from previous to current position
                    doc.Pages[pageNumber - 1].Graphics.DrawLine(blackPen,
                        new PdfPoint((double)m_previousX * conversionFactorX, (double)m_previousY * conversionFactorY),
                        new PdfPoint(currentX * conversionFactorX, currentY * conversionFactorY));

                    Windows.UI.Color penColor = new Windows.UI.Color();
                    Pen pen = new Pen(penColor);
                    pen.Color = Windows.UI.Colors.Black;

                    m_previousX = currentX;
                    m_previousY = currentY;
                    //refresh the view on screen
                    doc.Save(saveStream);
                    saveStream.Position = 0;
                    pdfViewer.LoadDocument(saveStream);
                    pdfViewer.Visibility = Windows.UI.Xaml.Visibility.Visible;

                    e.Handled = true;
                    //STEP 4: Save the file in a different name
                    //StorageFolder folder = KnownFolders.DocumentsLibrary;
                    //StorageFile saveFile;
                    //bool exists;
                    //try
                    //{
                    //    saveFile = await folder.GetFileAsync("TestDraw.pdf");
                    //    await pdfViewer.SaveDocumentAsync(saveFile);
                    //    exists = true;
                    //}
                    //catch
                    //{
                    //    //continue
                    //    exists = false;
                    //}
                    //if (!exists)
                    //{
                    //    try
                    //    {
                    //        saveFile = await folder.CreateFileAsync("TestDraw.pdf");
                    //        await pdfViewer.SaveDocumentAsync(saveFile);
                    //    }
                    //    catch
                    //    {
                    //        //just continue
                    //    }
                    //}

                    //Pen pen1 = new System.Drawing.Pen(Color.Black, 3);
                    //Pen pen2 = new System.Drawing.Pen(panel1.BackColor, 3);
                    //Graphics g = panel1.CreateGraphics();
                    //g.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.HighQuality;
                    //// Clear the graphics, creating a blank area to draw on
                    //g.Clear(panel1.BackColor);
                    //g.DrawLine(pen1, new Point(0, 0), new Point(e.Location.X, e.Location.Y));
                    //x_previous = e.Location.X;
                    //y_previous = e.Location.Y;
                }
            }
        }

        private void pdfViewer_PointerPressed(object sender, PointerRoutedEventArgs e)
        {
            Windows.UI.Input.PointerPoint ptrPoint = e.GetCurrentPoint(pdfViewer);

            m_previousX = ptrPoint.Position.X;
            m_previousY = ptrPoint.Position.Y;
            m_isTracing = true;
            System.Diagnostics.Debug.WriteLine("Start Tracing");
        }

        private void pdfViewer_PointerReleased(object sender, PointerRoutedEventArgs e)
        {
            m_isTracing = false;
            System.Diagnostics.Debug.WriteLine("End Tracing");
        }

        private async void DrawLineOnPDF()
        {
            //draw the line on the PDF loaded in the memory, based on the lines drawn on the canvas
            PdfPen blackPen = new PdfPen(new PdfRgbColor(100, 100, 100), 5);
            //get the loaded document from the viewer by saving it into a stream
            var memStream = new MemoryStream();
            Stream saveStream = memStream.AsRandomAccessStream().AsStreamForWrite();
            await pdfViewer.SaveDocumentAsync(saveStream);

            //get the current page no
            int pageNumber = pdfViewer.PageNumber;
            //load stream into an editable document
            PdfFixedDocument doc = new PdfFixedDocument(saveStream);
            //get the selected page
            if (pageNumber - 1 < 0) { pageNumber = 1; }
            //get the scaling factor to convert position on page to position on the PDF page
            double pageWidth = doc.Pages[pageNumber - 1].Width;
            double pageHeight = doc.Pages[pageNumber - 1].Height;
            double conversionFactorX = pageWidth / pdfViewer.ActualWidth;
            double conversionFactorY = pageHeight / pdfViewer.ActualHeight;
            //draw lines on the selected page, based on the lines drawn in the canvas
            foreach (ViewerPoint curPoint in ViewerPointsList)
            {
                double startX, startY, finishX, finishY;
                startX = curPoint.X1 * conversionFactorX;
                startY = curPoint.Y1 * conversionFactorY;
                finishX = curPoint.X2 * conversionFactorX;
                finishY = curPoint.Y2 * conversionFactorY;
                //draw a line on the selected page from previous to current position
                doc.Pages[pageNumber - 1].Graphics.DrawLine(blackPen,
                    new PdfPoint(startX, startY),
                    new PdfPoint(finishX, finishY));
            }
            //finally, clear the points list
            ViewerPointsList.Clear();
        }

        #endregion

        #region "FoxIT SDK Code"

        private async void OpenFileForFoxIt(string fileName)
        {
            StorageFolder folder = KnownFolders.DocumentsLibrary;
            StorageFile file = await folder.GetFileAsync(fileName);
            if (file == null)
                return;

            m_pdfFile = file;

            //load document from the file
            m_doc = await Document.LoadAsync(file, "");

            string fileID;
            fileID = file.FolderRelativeId;
            m_FileID = fileID;

            //load first page
            m_page = Document.LoadPage(m_doc, 0);
            if (m_page.pointer == 0)
                return;

            bool bRet = Document.StartParsingPage(m_page, false, null);
            //size up the PDF and the page
            //BBO: NEED TO PLAY AROUND TO SEE EXACTLY HOW THEY WORK
            m_ActualSizeX = (int)View.GetPageWidth(m_page);
            m_ActualSizeY = (int)View.GetPageHeight(m_page);

            if (!(m_screenHeight > 0))
                m_screenHeight = scroll_pdf.Height;
            float ratio = (float)scroll_pdf.Height / (float)m_ActualSizeY;

            m_ActualSizeX = (int)(m_ActualSizeX * ratio);
            m_ActualSizeY = (int)(m_ActualSizeY * ratio);

            img_pdfpage.Width = m_ActualSizeX;
            img_pdfpage.Height = m_ActualSizeY;

            Thickness thickness;
            thickness.Left = 5;
            thickness.Right = 5;
            thickness.Top = 0;
            thickness.Bottom = 0;

            img_pdfpage.Margin = thickness;
            m_pageRect.X = thickness.Left;
            m_pageRect.Width = m_ActualSizeX;
            m_pageRect.Y = 0;
            m_pageRect.Height = m_ActualSizeY;

            RenderFoxITPage(m_page, (int)m_ActualSizeX, (int)m_ActualSizeY);
            //initPsiEnv();

        }

        private async void RenderFoxITPage(PageHandle hPage, int width, int height)
        {
            //rendering page as a bitmap image
            m_bitmap = new PixelSource();
            m_bitmap.Width = width;
            m_bitmap.Height = height;
            await View.RenderPageAsync(m_bitmap, hPage, 0, 0, width, height, 0, (int)(Foxit.General.RenderFlags.Annot | Foxit.General.RenderFlags.LCDText), null);

            if (null == m_bytes)
                m_bytes = new byte[m_bitmap.PixelBuffer.Length];
            if (null == m_wb)
                m_wb = new WriteableBitmap(m_bitmap.Width, m_bitmap.Height);

            RenderFoxITDIB();

            img_pdfpage.Source = m_wb;

        }

        private void RenderFoxITDIB()
        {
            DataReader reader = DataReader.FromBuffer(m_bitmap.PixelBuffer);
            reader.ReadBytes(m_bytes);

            Stream stream = m_wb.PixelBuffer.AsStream();
            stream.Write(m_bytes, 0, m_bytes.Length);
        }
        private void RegisterForShare()
        {
            //BBO: not sure what this is for, no need to do any shares now
            DataTransferManager dataTransferManager = DataTransferManager.GetForCurrentView();
            dataTransferManager.DataRequested += new TypedEventHandler<DataTransferManager,
            DataRequestedEventArgs>(this.ShareStorageItemsHandler);
        }

        private void ShareStorageItemsHandler(DataTransferManager sender, DataRequestedEventArgs e)
        {
            //BBO: not sure what this is for, no need to do any shares now
            if (m_pdfFile == null)
                return;
            DataRequest request = e.Request;
            request.Data.Properties.Title = "Share PDF Files";
            request.Data.Properties.Description = "Share PDF with your friends.";
            // Because we are making async calls in the DataRequested event handler,
            // we need to get the deferral first.
            DataRequestDeferral deferral = request.GetDeferral();

            // Make sure we always call Complete on the deferral.
            try
            {
                List<IStorageItem> storageItems = new List<IStorageItem>();
                storageItems.Add(m_pdfFile);
                request.Data.SetStorageItems(storageItems);
            }
            finally
            {
                deferral.Complete();
            }

        }

        public void RenderPsi(int left, int top, int right, int bottom)
        {
            if (left < 0 && right < 0)
                return;
            if (top < 0 && bottom < 0)
                return;

            Rect rect = new Rect(left, top, right - left, bottom - top);
            Rect pageRect = new Rect(0, 0, m_pageRect.Width, m_pageRect.Height);
            rect.Intersect(pageRect);
            if (rect.IsEmpty || 0 == rect.Width || 0 == rect.Height)
                return;

            PSI.Render(m_hPSI, m_bitmap, (int)rect.X, (int)rect.Y, (int)rect.Width, (int)rect.Height, (int)rect.X, (int)rect.Y);
            if (0 == m_unionRect.Width || 0 == m_unionRect.Height)
                m_unionRect = rect;
            m_unionRect.Union(rect);

            DataReader reader = DataReader.FromBuffer(m_bitmap.PixelBuffer);
            reader.ReadBytes(m_bytes);

            Stream stream = m_wb.PixelBuffer.AsStream();
            //update the psi rect
            int pitch = (m_bitmap.Width * 8 * 4 + 31) / 32 * 4;
            for (int i = top; i < bottom; i++)
            {
                stream.Seek(i * pitch + left * 4, SeekOrigin.Begin);
                stream.Write(m_bytes, i * pitch + left * 4, (right - left) * 4);
            }

            m_wb.Invalidate();
        }

        private void Grid_PointerPressed(object sender, PointerRoutedEventArgs e)
        {

        }

        private void Grid_PointerMoved(object sender, PointerRoutedEventArgs e)
        {

        }

        private void Grid_PointerReleased(object sender, PointerRoutedEventArgs e)
        {

        }

        private void scroll_pdf_ViewChanged(object sender, ScrollViewerViewChangedEventArgs e)
        {

        }

        #endregion

        private void inkCanvas_PointerPressed(object sender, PointerRoutedEventArgs e)
        {
            // Get information about the pointer location.

            PointerPoint pt = e.GetCurrentPoint(inkCanvas);
            m_PreviousContactPoint = pt.Position;

            // Accept input only from a pen or mouse with the left button pressed.
            PointerDeviceType pointerDevType = e.Pointer.PointerDeviceType;
            if (pointerDevType == PointerDeviceType.Pen ||
                pointerDevType == PointerDeviceType.Mouse && pt.Properties.IsLeftButtonPressed)
            {
                m_PenId = pt.PointerId;
                e.Handled = true;
            }
            else if (pointerDevType == PointerDeviceType.Touch)
            {
                // Process touch input
            }
        }

        private double CalculatedDistance(double x1, double y1, double x2, double y2)
        {
            double d = 0;
            d = Math.Sqrt(Math.Pow((x2 - x1), 2) + Math.Pow((y2 - y1), 2));
            return d;
        }

        public void inkCanvas_PointerReleased(object sender, PointerRoutedEventArgs e)
        {
            if (e.Pointer.PointerId == m_PenId)
            {
                PointerPoint pt = e.GetCurrentPoint(inkCanvas);

                //invoke the drawing on PDF
                DrawLineOnPDF();

                if (m_CurrentMode == "Erase")
                {
                    System.Diagnostics.Debug.WriteLine("Erasing : Pointer Released");

                    m_InkManager.ProcessPointerUp(pt);
                    m_HighLightManager.ProcessPointerUp(pt);
                }
                else
                {
                    // Pass the pointer information to the InkManager. 
                    //CurrentManager.ProcessPointerUp(pt);
                }
            }
            else if (e.Pointer.PointerId == _touchID)
            {
                // Process touch input 
            }

            _touchID = 0;
            m_PenId = 0;

            // Call an application-defined function to render the ink strokes. 
            //RefreshCanvas();

            e.Handled = true;
        }

        private void inkCanvas_PointerMoved(object sender, PointerRoutedEventArgs e)
        {
            if (e.Pointer.PointerId == m_PenId)
            {
                PointerPoint pt = e.GetCurrentPoint(inkCanvas);

                currentContactPt = pt.Position;
                x1 = m_PreviousContactPoint.X;
                y1 = m_PreviousContactPoint.Y;
                x2 = currentContactPt.X;
                y2 = currentContactPt.Y;

                var color = m_CurrentDrawingColor;
                var size = m_CurrentDrawingSize;

                if (CalculatedDistance(x1, y1, x2, y2) > 2.0)
                {
                    //
                    // If the delta of the mouse is significant enough,
                    // we add a line geometry to the Canvas
                    Line line = new Line()
                    {
                        X1 = x1,
                        Y1 = y1,
                        X2 = x2,
                        Y2 = y2,
                        StrokeThickness = size,
                        Stroke = new SolidColorBrush(color)
                    };

                    //add the line values to the list, so they will be drawn by the PDF object
                    ViewerPoint curPoint = new ViewerPoint();
                    curPoint.X1 = x1;
                    curPoint.Y1 = y1;
                    curPoint.X2 = x2;
                    curPoint.Y2 = y2;
                    ViewerPointsList.Add(curPoint);

                    m_PreviousContactPoint = currentContactPt;

                    // Draw the line on the canvas by adding the Line object as
                    // a child of the Canvas object.
                    inkCanvas.Children.Add(line);
                }
            }
        }

        private void btnInkingToggle_Click(object sender, RoutedEventArgs e)
        {
            switch(pdfViewer.RenderMode)
            {
                case C1.Xaml.PdfViewer.RenderMode.Default:
                    pdfViewer.RenderMode = C1.Xaml.PdfViewer.RenderMode.System;
                    break;
                case C1.Xaml.PdfViewer.RenderMode.System:
                    pdfViewer.RenderMode = C1.Xaml.PdfViewer.RenderMode.Default;
                    break;
            }
        }

        private void inkCanvas_SizeChanged(object sender, SizeChangedEventArgs e)
        {
            //pdfViewer.Height = inkCanvas.Height;
            //pdfViewer.Width = inkCanvas.Width;
        }
    }
}
