﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Runtime.InteropServices.WindowsRuntime;

using Windows.ApplicationModel.DataTransfer;
using Windows.Devices.Input;
using Windows.Foundation;
using Windows.Foundation.Collections;
using Windows.Graphics.Imaging;
using Windows.UI.Input;
using Windows.UI.Input.Inking;
using Windows.Storage;
using Windows.Storage.Streams;
using Windows.Storage.Provider;
using Windows.UI;
using Windows.UI.Xaml;
using Windows.UI.Xaml.Controls;
using Windows.UI.Xaml.Controls.Primitives;
using Windows.UI.Xaml.Data;
using Windows.UI.Xaml.Input;
using Windows.UI.Xaml.Media;
using Windows.UI.Xaml.Media.Imaging;
using Windows.UI.Xaml.Navigation;
using Windows.UI.Xaml.Shapes;
using Windows.UI.Popups;
using Windows.Storage.Pickers;

using Syncfusion.Windows.PdfViewer;
using Syncfusion.Pdf.Parsing;
using Syncfusion.Pdf.Graphics;
using System.Reflection;
using Syncfusion.Pdf.Interactive;
using Syncfusion.Pdf;

// The Blank Page item template is documented at http://go.microsoft.com/fwlink/?LinkId=234238

namespace PDFTest
{
    /// <summary>
    /// An empty page that can be used on its own or navigated to within a Frame.
    /// </summary>
    public sealed partial class SyncfusionViewerTestPage : Page
    {
        //form a struct and list of those to trace the lines drawn on the screen
        private struct ViewerPoint
        {
            public double X1;
            public double Y1;
            public double X2;
            public double Y2;
        }
        private List<ViewerPoint> ViewerPointsList = new List<ViewerPoint>();
        DispatcherTimer t;
        DispatcherTimer t_Test;
        PdfLoadedAnnotationCollection annots = null;
        System.Drawing.RectangleF[] annotationBounds;
        //Canvas related variables
        private Point m_PreviousContactPoint;
        private Point currentContactPt;
        private double x1;
        private double y1;
        private double x2;
        private double y2;
        private Color m_CurrentDrawingColor = Colors.Yellow;
        private Color m_CurrentHighlightColor = Colors.Yellow;
        //syncfusion related variables
        PdfUnitConvertor m_converter = new PdfUnitConvertor();
        int currentPageIndex;
        PdfLoadedDocument m_loadedDocument;
        int m_zoomFactor = 100;

        public SyncfusionViewerTestPage()
        {
            this.InitializeComponent();
            this.PointerPressed += SyncfusionViewerTestPage_PointerPressed;
            this.Holding += SyncfusionViewerTestPage_Holding;
            t = new DispatcherTimer();
            t.Tick += t_Tick;
            t.Interval = new TimeSpan(0, 0, 1);
            t.Start();
        }

        void SyncfusionViewerTestPage_PointerPressed(object sender, PointerRoutedEventArgs e)
        {
            PointerPoint pointerPosition = e.GetCurrentPoint(annotCanvas);
            Point point = pointerPosition.Position;
            ClickStarted(point);
            ClickReleased();
        }

        void SyncfusionViewerTestPage_Holding(object sender, HoldingRoutedEventArgs e)
        {
            if (e.HoldingState == Windows.UI.Input.HoldingState.Started)
            {
                Point pointerPosition = e.GetPosition(annotCanvas);
                ClickStarted(pointerPosition);
            }
            else if(e.HoldingState == Windows.UI.Input.HoldingState.Completed)
            {
                ClickReleased();
            }
        }

        private void ClickReleased()
        {
            float startX = 0.0f, startY = 0.0f, finishX = 0.0f, finishY = 0.0f, currentLocationY1 = 0.0f, currentLocationY2 = 0.0f;
            float zoomConversionFactor = syncFusionPDFViewer.Zoom / 100.0f;
            double pageWidth = m_loadedDocument.Pages[currentPageIndex].Size.Width * zoomConversionFactor;
            double pageHeight = m_loadedDocument.Pages[currentPageIndex].Size.Height * zoomConversionFactor;
            float pixelPageHeight = m_converter.ConvertToPixels((float)pageHeight, PdfGraphicsUnit.Point);
            float pixelPageWidth = m_converter.ConvertToPixels((float)pageWidth, PdfGraphicsUnit.Point);
            double pageSize = syncFusionPDFViewer.PageOffsetCollection[currentPageIndex + 1];
            int pageIndex = currentPageIndex;
            float OffsetPixelsX, OffsetPixelsY;
            if (ActualWidth > pixelPageWidth)
            {
                OffsetPixelsX = ((float)syncFusionPDFViewer.ActualWidth - pixelPageWidth);
                OffsetPixelsX = (OffsetPixelsX / 2.0f);
            }
            else
            {
                OffsetPixelsX = (zoomConversionFactor) - (float)syncFusionPDFViewer.HorizontalOffset;
            }
            if (syncFusionPDFViewer.VerticalOffset < pixelPageHeight)
            {
                OffsetPixelsY = (float)syncFusionPDFViewer.VerticalOffset - (syncFusionPDFViewer.PageGap * zoomConversionFactor);
            }
            else
            {
                OffsetPixelsY = (currentPageIndex * (((float)syncFusionPDFViewer.VerticalOffset / (currentPageIndex)) - pixelPageHeight)) - (syncFusionPDFViewer.PageGap * (currentPageIndex + 1) * zoomConversionFactor);
            }
            foreach (ViewerPoint curPoint in ViewerPointsList)
            {
                ViewerPoint viewPoint = curPoint;
                int currentPageNumber = 0;
                
                currentLocationY1 = (syncFusionPDFViewer.VerticalOffset + (float)viewPoint.Y1) / zoomConversionFactor;
                currentLocationY2 = (syncFusionPDFViewer.VerticalOffset + (float)viewPoint.Y2) / zoomConversionFactor;

                for (int i = 1; i <= syncFusionPDFViewer.PageCount; i++)
                {

                    currentPageNumber = (int)Math.Floor(currentLocationY1 / (pixelPageHeight / zoomConversionFactor));

                    if (currentPageNumber > 0 && (currentPageNumber < syncFusionPDFViewer.PageCount - 1 || syncFusionPDFViewer.PageCount == 2))
                    {
                        if (currentLocationY1 > syncFusionPDFViewer.PageOffsetCollection[currentPageNumber])
                        {
                            viewPoint.Y1 = (currentLocationY1 - syncFusionPDFViewer.PageGap) - syncFusionPDFViewer.PageOffsetCollection[currentPageNumber];
                            viewPoint.Y2 = (currentLocationY2 - syncFusionPDFViewer.PageGap) - syncFusionPDFViewer.PageOffsetCollection[currentPageNumber];
                            startX = m_converter.ConvertFromPixels(((float)viewPoint.X1 - (float)OffsetPixelsX) / zoomConversionFactor, PdfGraphicsUnit.Point);
                            finishX = m_converter.ConvertFromPixels(((float)viewPoint.X2 - (float)OffsetPixelsX) / zoomConversionFactor, PdfGraphicsUnit.Point);
                            startY = m_converter.ConvertFromPixels((pixelPageHeight / zoomConversionFactor) - (float)viewPoint.Y1, PdfGraphicsUnit.Point);
                            finishY = m_converter.ConvertFromPixels((pixelPageHeight / zoomConversionFactor) - (float)viewPoint.Y2, PdfGraphicsUnit.Point);
                            currentPageIndex = currentPageNumber;
                        }
                    }
                   
                    else
                    {
                        startX = m_converter.ConvertFromPixels(((float)viewPoint.X1 - (float)OffsetPixelsX), PdfGraphicsUnit.Point);
                        finishX = m_converter.ConvertFromPixels(((float)viewPoint.X2 - (float)OffsetPixelsX), PdfGraphicsUnit.Point);
                        startY = m_converter.ConvertFromPixels(pixelPageHeight - ((float)OffsetPixelsY + (float)viewPoint.Y1), PdfGraphicsUnit.Point);
                        finishY = m_converter.ConvertFromPixels(pixelPageHeight - ((float)OffsetPixelsY + (float)viewPoint.Y2), PdfGraphicsUnit.Point);
                        if (currentPageNumber == 0)
                        {
                            currentPageIndex = currentPageNumber;
                        }
                        startX = startX / zoomConversionFactor;
                        startY = startY / zoomConversionFactor;
                        finishX = finishX / zoomConversionFactor;
                        finishY = finishY / zoomConversionFactor;
                    }
                    break;
                }
            }

            ViewerPointsList.Clear();
            CheckPopupAnnotation(startX, startY, finishX, finishY, currentLocationY1);
        }

        private void ClickStarted(Point pointerPosition)
        {
            x1 = m_PreviousContactPoint.X;
            y1 = m_PreviousContactPoint.Y;
            x2 = pointerPosition.X;
            y2 = pointerPosition.Y;

            if (CalculatedDistance(x1, y1, x2, y2) > 2.0)
            {
                Line line = new Line()
                {
                    X1 = x1,
                    Y1 = y1,
                    X2 = x2,
                    Y2 = y2,
                };

                ViewerPoint curPoint = new ViewerPoint();
                curPoint.X1 = x1;
                curPoint.Y1 = y1;
                curPoint.X2 = x2;
                curPoint.Y2 = y2;
                ViewerPointsList.Add(curPoint);

                m_PreviousContactPoint = currentContactPt;
            }
        }

        private void CheckPopupAnnotation(float startX, float startY, float finishX, float finishY,float currentLocationY1)
        {
            float popupAnnotXPosition = finishX;
            float popupAnnotYPosition = startY - finishY;
            for(int i=0;i<annotationBounds.Length;i++)
            {
                float minXBounds = annotationBounds[i].X;
                float minYBounds = annotationBounds[i].Y;
                float maxXBounds = annotationBounds[i].X + annotationBounds[i].Width;
                float maxYBounds = annotationBounds[i].Y + annotationBounds[i].Height;
                if ((popupAnnotYPosition <= minYBounds && popupAnnotYPosition <= maxYBounds))
                {
                    if (currentLocationY1 > minYBounds)
                        popupAnnotYPosition = currentLocationY1;
                    else
                        popupAnnotYPosition = currentLocationY1 + popupAnnotYPosition;
                        if(popupAnnotYPosition>maxYBounds)
                        {
                            popupAnnotYPosition = maxYBounds;
                        }
                }
                if((popupAnnotXPosition>=minXBounds&&popupAnnotXPosition<=maxXBounds)&&(popupAnnotYPosition>=minYBounds&&popupAnnotYPosition<=maxYBounds))
                {
                    MessageDialog msgDialog = new MessageDialog("Popup Annotation Clicked");
                    msgDialog.ShowAsync();
                }
            }
        }

        #region "Loading Methods"

        private void t_Tick(object sender, object e)
        {
            string fileName = "a.pdf";
            OpenFile(fileName);
            t.Stop();
        }

        private void OpenFile(string fileName)
        {
            Assembly assembly = typeof(SyncfusionViewerTestPage).GetTypeInfo().Assembly;
            Stream fileStream = assembly.GetManifestResourceStream("PDFTest.Assets.a.pdf");
            m_loadedDocument = new PdfLoadedDocument(fileStream);
            for (int i = 0; i < m_loadedDocument.Pages.Count;i++ )
            {
                annots = m_loadedDocument.Pages[i].Annotations;
            }
            RenderPopupBounds(annots);
            syncFusionPDFViewer.LoadDocument(m_loadedDocument);
            syncFusionPDFViewer.ViewMode = PageViewMode.FitWidth;
            currentPageIndex = 0;
            t_Test = new DispatcherTimer();
            t_Test.Interval = new TimeSpan(0, 0, 1);
            t_Test.Start();
        }

        private void RenderPopupBounds(PdfLoadedAnnotationCollection annots)
        {
          annotationBounds = new System.Drawing.RectangleF[annots.Count];
          for(int i=0;i<annots.Count;i++)
          {
              PdfLoadedPopupAnnotation annotPopup = annots[i] as PdfLoadedPopupAnnotation;
              if (annotPopup != null)
              {
                  annotationBounds[i] = (annotPopup).Bounds;
              }
          }
        }

        #endregion

        private double CalculatedDistance(double x1, double y1, double x2, double y2)
        {
            double d = 0;
            d = Math.Sqrt(Math.Pow((x2 - x1), 2) + Math.Pow((y2 - y1), 2));
            return d;
        }
    }
}
